"""This module is a generic for all pkpass commands"""
import getpass
from os import getcwd, path, sep
from sqlalchemy.orm import sessionmaker
from libpkpass import LOGGER
from libpkpass.commands.arguments import ARGUMENTS as arguments
from libpkpass.crypto import print_card_info
from libpkpass.errors import (
    CliArgumentError,
    PasswordIOError,
    NotThePasswordOwnerError,
)
from libpkpass.identities import IdentityDB
from libpkpass.passworddb import PasswordDB
from libpkpass.password import PasswordEntry
from libpkpass.util import collect_args, color_prepare, build_recipient_list
from libpkpass.models.recipient import Recipient


class Command:
    ##########################################################################
    """Base class for all commands.  Auotmatically registers with cli subparser
    and provides run execution for itself."""
    ##########################################################################
    name = None
    description = None
    selected_args = [
        "nocache",
        "verbosity",
        "quiet",
        "identity",
        "cabundle",
        "certpath",
        "color",
        "theme_map",
    ]
    passphrase = None

    def __init__(self, cli, iddb=None, pwdb=None):
        ##################################################################
        """Intialization function for class. Register with argparse"""
        ##################################################################
        # default certpath to none because connect string is allowed
        self.args = {}
        self.recipient_list = []
        self.escrow_and_recipient_list = []
        self.iddb = iddb if iddb else IdentityDB()
        self.pwdbcached = pwdb is not None
        self.passworddb = pwdb if pwdb else PasswordDB()
        cli.register(self, self.name, self.description)

    def register(self, parser):
        ##################################################################
        """Registration function for class. Register with argparse"""
        ##################################################################
        for arg in sorted(self.selected_args):
            parser.add_argument(*arguments[arg]["args"], **arguments[arg]["kwargs"])

    def run(self, parsedargs):
        ##################################################################
        """Passes the argparse Namespace object of parsed arguments"""
        ##################################################################
        self._run_command_setup(parsedargs)
        self._pre_check()
        self._passphrase_check()
        return self._run_command_execution()

    def _passphrase_check(self):
        if "nopassphrase" in self.selected_args and not self.args["nopassphrase"]:
            for mesg in print_card_info(
                self.args["card_slot"],
                self.iddb.id,
                self.args["verbosity"],
                self.args["color"],
                self.args["theme_map"],
                self.args["SCBackend"],
            ):
                LOGGER.info(mesg)
            self.passphrase = getpass.getpass("Enter Pin/Passphrase: ")

    def _run_command_setup(self, parsedargs):
        ##################################################################
        """Passes the argparse Namespace object of parsed arguments"""
        ##################################################################
        self.args = collect_args(parsedargs)
        self.iddb.session = sessionmaker(bind=self.args["db"]["engine"])()
        self._validate_combinatorial_args()
        self._validate_args()

        # Build the list of recipients that this command will act on
        self.recipient_list, self.escrow_and_recipient_list = build_recipient_list(self.args)

        # If there are defined repositories of keys and certificates, load them
        self.iddb.args = self.args
        self.iddb.cabundle = self.args["cabundle"]
        self.iddb.load_certs_from_directory(
            self.args["certpath"],
            connectmap=self.args["connect"],
            nocache=self.args["no_cache"],
        )
        if self.args["keypath"]:
            self.iddb.load_keys_from_directory(self.args["keypath"])
        self.iddb.id = (
            self.iddb.session.query(Recipient)
            .filter(Recipient.name == self.args["identity"])
            .first()
        )
        self._validate_identities()
        self.iddb.id = dict(self.iddb.id)

        if (
            self.args["subparser_name"]
            in ["list", "interpreter", "distribute", "export"]
            and not self.pwdbcached
        ):
            self.passworddb.load_from_directory(self.args["pwstore"])
        if "pwname" in self.args and self.args["pwname"]:
            self._resolve_directory_path()
        self.args["card_slot"] = self.args["card_slot"] if self.args["card_slot"] else 0

    def _resolve_directory_path(self):
        ####################################################################
        """This handles how a user inputs the pwname, this tries to be smart
        good luck everybody else"""
        ####################################################################
        pwd = getcwd()
        pwd_pwname = path.normpath(path.join(pwd, self.args["pwname"]))
        if self.args["pwstore"] in pwd_pwname:
            self.args["pwname"] = pwd_pwname.replace(self.args["pwstore"] + sep, "")

    def safety_check(self):
        ##################################################################
        """This provides a sanity check that you are the owner of a password."""
        ##################################################################
        try:
            password = PasswordEntry()
            password.read_password_data(
                path.join(self.args["pwstore"], self.args["pwname"])
            )
            return (
                self.args["identity"] in password["recipients"],
                password["metadata"]["creator"],
            )
        except PasswordIOError:
            return (True, None)

    def update_pass(self, pass_value):
        ##################################################################
        """Fully updated a password record"""
        ##################################################################
        pass_entry = PasswordEntry()
        pass_entry.read_password_data(
            path.join(self.args["pwstore"], self.args["pwname"])
        )
        swap_pass = PasswordEntry()
        swap_pass.add_recipients(
            secret=pass_value,
            distributor=self.args["identity"],
            recipients=[self.args["identity"]],
            session=self.iddb.session,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            escrow_users=self.args["escrow_users"],
            minimum=self.args["min_escrow"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )
        pass_entry["recipients"][self.args["identity"]] = swap_pass["recipients"][
            self.args["identity"]
        ]
        pass_entry.write_password_data(
            path.join(self.args["pwstore"], self.args["pwname"]),
            overwrite=self.args["overwrite"],
        )

    def create_pass(self, password1, description, authorizer, recipient_list=None):
        ##################################################################
        """This writes password data to a file."""
        ##################################################################
        password_metadata = {}
        password_metadata["description"] = description
        password_metadata["authorizer"] = authorizer
        password_metadata["creator"] = self.args["identity"]
        password_metadata["name"] = self.args["pwname"]
        if self.args["noescrow"]:
            self.args["min_escrow"] = None
            self.args["escrow_users"] = None
        if recipient_list is None:
            recipient_list = [self.args["identity"]]

        password = PasswordEntry(**password_metadata)

        password.add_recipients(
            secret=password1,
            distributor=self.args["identity"],
            recipients=recipient_list,
            session=self.iddb.session,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            escrow_users=self.args["escrow_users"],
            minimum=self.args["min_escrow"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )

        password.write_password_data(
            path.join(self.args["pwstore"], self.args["pwname"]),
            overwrite=self.args["overwrite"],
        )

    def create_or_update_pass(
        self, password1, description, authorizer, recipient_list=None
    ):
        ##################################################################
        """This creates new or updates existing passwords"""
        ##################################################################
        safe, owner = self.safety_check()
        if safe or self.args["overwrite"]:
            if owner and owner != self.args["identity"]:
                self.update_pass(password1)
            else:
                self.create_pass(password1, description, authorizer, recipient_list)
        else:
            raise NotThePasswordOwnerError(
                self.args["identity"], owner, self.args["pwname"]
            )

    def _run_command_execution(self):
        ##################################################################
        """Passes the argparse Namespace object of parsed arguments"""
        ##################################################################
        raise NotImplementedError

    def _validate_args(self):
        raise NotImplementedError

    def _pre_check(self):
        """Pre check likely won't happen on all commands so we'll allow a
        pass"""

    def _validate_combinatorial_args(self):
        ##################################################################
        """This is a weird function name so: combinatorial in this case
        means that one of the 'combinatorial' arguments are required
        however, not all of them are necessarily required.
        Ex: We need certs, we can get this from certpath or connect
        we do not need both of these arguments but at least one is
        required"""
        ##################################################################
        # we want a multi-dim of lists, this way if more combinations come up
        # that would be required in a 1 or more capacity, we just add
        # a list to this list
        args_list = [["certpath", "connect"]]
        for arg_set in args_list:
            valid = False
            for arg in arg_set:
                if arg in self.args and self.args[arg] is not None:
                    valid = True
                    break
            if not valid:
                raise CliArgumentError(f"'{arg_set[0]}' or '{arg_set[1]}' is required")

    def _validate_identities(self, swap_list=None):
        ##################################################################
        """Ensure identities meet criteria for processing"""
        ##################################################################
        if not swap_list:
            swap_list = self.escrow_and_recipient_list
        for recipient in swap_list:
            self.iddb.verify_identity(recipient)
            if recipient not in [
                x[0] for x in self.iddb.session.query(Recipient.name).all()
            ]:
                raise CliArgumentError(
                    f"Error: Recipient '{recipient}' is not in the recipient database"
                )

        if self.iddb.id:
            self.iddb.verify_identity(self.args["identity"])
        else:
            raise CliArgumentError(
                f"Error: Your user '{self.args['identity']}' is not in the recipient database"
            )

    def color_print(self, string, color_type):
        ##################################################################
        """Handle the color printing for objects"""
        ##################################################################
        return color_prepare(
            string, color_type, self.args["color"], self.args["theme_map"]
        )



"""This is used to check the os requirements"""
from os import path
from shutil import which
from libpkpass.commands.command import Command


class VerifyInstall(Command):
    ####################################################################
    """This class is used as a command object and parses information passed through
    the CLI to show passwords that have been distributed to users"""
    ####################################################################
    name = "verifyinstall"
    description = "verify required software is present"
    selected_args = Command.selected_args + [
        "pwstore",
        "keypath",
        "noverify",
        "card_slot",
        "connect",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        yield from print_messages(check_required_software, "installed software check")
        yield from print_messages(
            check_passdb,
            "passdb check",
            cabundle=path.realpath(self.args["cabundle"]),
            pwstore=path.realpath(self.args["pwstore"]),
            keypath=path.realpath(self.args["keypath"])
            if self.args["keypath"]
            else None,
            certpath=path.realpath(self.args["certpath"])
            if self.args["certpath"]
            else None,
            connect=self.args["connect"],
        )

    def _validate_args(self):
        pass


def check_exists(name):
    ####################################################################
    """Check whether a program exists in path"""
    ####################################################################
    return which(name) is not None


def print_messages(func, msg, **kwargs):
    yield f"Starting {msg}"
    yield func(**kwargs)


def check_required_software():
    required_tools = {
        "pkcs15-tool (available via opensc)": ["pkcs15-tool"],
        "ssl (openssl or libressl)": ["openssl", "libressl"],
    }
    not_found = []
    for key, value in required_tools.items():
        found_tool = False
        for tool in value:
            if check_exists(tool):
                found_tool = True
        if not found_tool:
            not_found.append(key)
    if not_found:
        return "The following packages were not found: \n\t%s" % "\n\t".join(not_found)
    return "Successful installed software check"


def check_passdb(cabundle, pwstore, certpath=None, keypath=None, connect=None):
    ret_msg = []
    if not path.isfile(cabundle):
        ret_msg.append("Cabundle is not a file")
    if not path.isdir(pwstore):
        ret_msg.append("pwstore is not a directory")
    if connect and certpath:
        ret_msg.append("certpath or keypath is defined while using a connector")
    if certpath and not path.isdir(certpath):
        ret_msg.append(f"certpath is not a directory {certpath}")
    if keypath and not path.isdir(keypath):
        ret_msg.append(f"Keypath is not a directory: {keypath}")
    ret_msg.append("Completed passdb check")
    return "\n".join(ret_msg)



"""This module allows for the renaming of passwords"""
from os import path, rename
from sys import exit as sys_exit
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError, NotThePasswordOwnerError, PasswordIOError


class Rename(Command):
    ####################################################################
    """This class implements the CLI functionality of renaming of passwords"""
    ####################################################################

    name = "rename"
    description = "Rename a password in the repository"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "overwrite",
        "stdin",
        "nopassphrase",
        "keypath",
        "card_slot",
        "rename",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        safe, owner = self.safety_check()
        if safe and owner:
            orig_pass = self.args["pwname"]
            self.args["pwname"] = self.args["rename"]
            resafe, reowner = self.safety_check()
            self.args["pwname"] = orig_pass
            if resafe or self.args["overwrite"]:
                password = PasswordEntry()
                password.read_password_data(
                    path.join(self.args["pwstore"], self.args["pwname"])
                )
                plaintext_pw = password.decrypt_entry(
                    identity=self.iddb.id,
                    passphrase=self.passphrase,
                    card_slot=self.args["card_slot"],
                    SCBackend=self.args["SCBackend"],
                    PKCS11_module_path=self.args["PKCS11_module_path"],
                )
                self._confirmation(plaintext_pw)
            else:
                raise NotThePasswordOwnerError(
                    self.args["identity"], reowner, self.args["rename"]
                )
        else:
            raise NotThePasswordOwnerError(
                self.args["identity"], owner, self.args["pwname"]
            )
        # necessary for print statement
        yield ""

    def rename_pass(self):
        ##################################################################
        """This renames a password that the user has created"""
        ##################################################################
        oldpath = path.join(self.args["pwstore"], self.args["pwname"])
        newpath = path.join(self.args["pwstore"], self.args["rename"])
        try:
            rename(oldpath, newpath)
            password = PasswordEntry()
            password.read_password_data(newpath)
            password["metadata"]["name"] = self.args["rename"]
            password.write_password_data(newpath)

        except OSError as err:
            raise PasswordIOError(
                f"Password '{self.args['pwname']}' not found"
            ) from err

    def _confirmation(self, plaintext_pw):
        ####################################################################
        """Run confirmation for rename"""
        ####################################################################
        yes = {"yes", "y", "ye", ""}
        deny = {"no", "n"}
        confirmation = input(
            f"{self.args['pwname']}: {plaintext_pw}\nRename this password?(Defaults yes): "
        )
        if confirmation.lower() in yes:
            self.rename_pass()
        elif confirmation.lower() in deny:
            sys_exit()
        else:
            print("please respond with yes or no")
            self._confirmation(plaintext_pw)

    def _validate_args(self):
        ####################################################################
        """Validate necessary arguments"""
        ####################################################################
        for argument in ["pwname", "keypath", "rename"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _pre_check(self):
        if path.exists(path.join(self.args["pwstore"], self.args["pwname"])):
            return True
        raise PasswordIOError(
            f"{path.join(self.args['pwstore'], self.args['pwname'])} does not exist"
        )



"""This module allows for the population of external password stores"""
from os import path, remove
from re import finditer
from base64 import standard_b64encode
from ruamel.yaml import YAML
from yaml import dump
from libpkpass import LOGGER
from libpkpass.commands.show import Show
from libpkpass.crypto import puppet_password
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError
from libpkpass.models.recipient import Recipient


class Populate(Show):
    ####################################################################
    """This class implements the CLI functionality of password store integration"""
    ####################################################################

    name = "populate"
    description = (
        "Populate external resource with password, currently supports: puppet_eyaml"
    )
    selected_args = [
        "pwname",
        "cabundle",
        "card_slot",
        "nopassphrase",
        "all",
        "certpath",
        "color",
        "identity",
        "keypath",
        "nocache",
        "noverify",
        "pwstore",
        "quiet",
        "theme_map",
        "type",
        "value",
        "verbosity",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        # currently only support puppet
        pop_type = self.args["type"]
        if pop_type not in ["puppet_eyaml", "kubernetes"]:
            raise CliArgumentError(f"'{pop_type}' is an unsupported population type")

        if pop_type == "kubernetes":
            yield from self._handle_kubernetes()
        elif self.args["all"]:
            for password in self.args["populate"][pop_type]["passwords"].keys():
                pwentry = PasswordEntry()
                self._decrypt_wrapper(self.args["pwstore"], pwentry, password, pop_type)

        elif (
            self.args["pwname"]
            not in self.args["populate"][pop_type]["passwords"].keys()
        ):
            raise CliArgumentError(
                f"'{self.args['pwname']}' doesn't have a mapping in {pop_type}"
            )
        else:
            password = PasswordEntry()
            yield from self._decrypt_wrapper(
                self.args["pwstore"], password, self.args["pwname"], pop_type
            )

    def _handle_kubernetes(self):
        ####################################################################
        """This function creates a file containing kube secrets"""
        ####################################################################
        k_conf = self.args["populate"]["kubernetes"]
        plaintext = {}
        for password in k_conf["needed"]:
            pwentry = PasswordEntry()
            pwentry.read_password_data(path.join(self.args["pwstore"], password))
            plaintext_pw = self._decrypt_password_entry(pwentry)
            plaintext[password] = plaintext_pw
        if "output" in k_conf:
            if self.args["pwstore"] in k_conf["output"]:
                raise CliArgumentError(
                    "Kubernetes output file should not exist in password store"
                )
            if path.isfile(k_conf["output"]):
                remove(k_conf["output"])
        for args in k_conf["passwords"]:
            data = self._kubernetes_match_loop(args, plaintext)
            kube_map = {
                "kind": "Secret",
                "apiVersion": args["apiVersion"],
                "metadata": args["metadata"],
                "data": data,
                "type": args["type"],
            }
            if self.args["value"] or "output" not in k_conf:
                yield f"---\n{dump(kube_map)}"
            else:
                with open(k_conf["output"], "a", encoding="ASCII") as fname:
                    print(f"---\n{dump(kube_map)}", file=fname)

    def _kubernetes_match_loop(self, args, plaintext):
        ####################################################################
        """This matches ${example} inside the arg defs and generates a
        dictionary of b64 encoded values with the actual password values
        placed instead of ${}"""
        ####################################################################
        data = {}
        for key, value in args["data"].items():
            new_pw = str(value)
            match_iter = finditer(r"\$\{([^${]*)\}", new_pw)
            if not match_iter:
                data[key] = standard_b64encode(new_pw.encode("ASCII")).decode()
            else:
                for match_values in match_iter:
                    match_value = match_values.group(1)
                    new_pw = new_pw.replace(
                        f"%{{{match_value}}}", plaintext[match_value]
                    )
                data[key] = standard_b64encode("".join(new_pw).encode("ASCII")).decode()
        return data

    def _handle_puppet(self, plaintext_pw, pwname):
        directory = path.expanduser(self.args["populate"]["puppet_eyaml"]["directory"])
        if not path.isdir(directory):
            raise CliArgumentError(f"'{directory}' is not a directory")
        puppet_bin = path.expanduser(self.args["populate"]["puppet_eyaml"]["bin"])
        if not path.isfile(puppet_bin):
            raise CliArgumentError(f"'{puppet_bin}' is not a file")
        pkcs7_pass = puppet_password(puppet_bin, plaintext_pw)
        if self.args["value"]:
            yield pkcs7_pass
        else:
            for hiera_file, names in self.args["populate"]["puppet_eyaml"]["passwords"][
                pwname
            ].items():
                with open(
                    path.join(directory, hiera_file), "r", encoding="ASCII"
                ) as data_file:
                    yield f"Updating: {hiera_file}"
                    yaml = YAML()
                    yaml.indent(mapping=2, sequence=4, offset=2)
                    yaml.preserve_quotes = True
                    hiera_yaml = yaml.load(data_file.read())
                    for name in names:
                        yield f"Updating: {name}"
                        hiera_yaml[name] = pkcs7_pass
                with open(
                    path.join(directory, hiera_file), "w", encoding="ASCII"
                ) as data_file:
                    yaml.dump(hiera_yaml, data_file)

    def _validate_args(self):
        for argument in ["keypath", "type"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _decrypt_wrapper(self, directory, password, pwname, pop_type):
        ####################################################################
        """Decide whether to decrypt normally or for escrow"""
        ####################################################################
        password.read_password_data(path.join(directory, pwname))
        plaintext_pw = self._decrypt_password_entry(password)
        if pop_type == "puppet_eyaml":
            yield from self._handle_puppet(plaintext_pw, pwname)

    def _decrypt_password_entry(self, password):
        ####################################################################
        """This decrypts a given password entry"""
        ####################################################################
        plaintext_pw = password.decrypt_entry(
            identity=self.iddb.id,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )
        distributor = password.recipients[self.iddb.id["name"]]["distributor"]
        if not self.args["noverify"]:
            result = password.verify_entry(
                self.iddb.id["name"],
                self.iddb,
                distributor,
                self.iddb.session.query(Recipient)
                .filter(Recipient.name == distributor)
                .first()
                .certs,
            )
            if not result["sigOK"]:
                LOGGER.warning(
                    "Could not verify that %s correctly signed your password entry.",
                    result["distributor"],
                )
            if not result["certOK"]:
                LOGGER.warning(
                    "Could not verify the certificate authenticity of user '%s'.",
                    result["distributor"],
                )
        return plaintext_pw



"""This Module allows for editing metadata of passwords"""
from os import path
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError


class Modify(Command):
    ####################################################################
    """This class implements the cli list"""
    ####################################################################
    name = "modify"
    description = "Modify the metadata of a password"
    selected_args = Command.selected_args + ["pwname", "pwstore"]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        full_path = path.join(self.args["pwstore"], self.args["pwname"])
        password = PasswordEntry()
        password.read_password_data(full_path)
        editable = ["authorizer", "description"]
        for key, value in password["metadata"].items():
            if key in editable:
                yield f"""{self.color_print(f"Current value for '{key}':", "first_level")} {value}"""
                password["metadata"][key] = input(f"New Value for {key}: ")

        password.write_password_data(full_path)

    def _validate_args(self):
        for argument in ["pwname", "pwstore"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



"""This Module allows for copying a password directly to the clipboard"""
from time import sleep
from os import path
from pyperclip import copy, paste
from libpkpass import LOGGER
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError, PasswordIOError
from libpkpass.models.recipient import Recipient


class Clip(Command):
    ####################################################################
    """This class allows for the copying of a password to the clipboard"""

    ####################################################################
    name = "clip"
    description = "Copy a password to clipboard"
    selected_args = Command.selected_args + [
        "pwname",
        "pwstore",
        "stdin",
        "time",
        "keypath",
        "nopassphrase",
        "noverify",
        "card_slot",
    ]

    def _run_command_execution(self):
        ####################################################################
        """Run function for class."""
        ####################################################################
        password = PasswordEntry()
        password.read_password_data(
            path.join(self.args["pwstore"], self.args["pwname"])
        )
        distributor = password.recipients[self.iddb.id["name"]]["distributor"]
        plaintext_pw = password.decrypt_entry(
            identity=self.iddb.id,
            passphrase=self.passphrase,
            card_slot=self.args["card_slot"],
            SCBackend=self.args["SCBackend"],
            PKCS11_module_path=self.args["PKCS11_module_path"],
        )
        if not self.args["noverify"]:
            result = password.verify_entry(
                self.iddb.id["name"],
                self.iddb,
                distributor,
                self.iddb.session.query(Recipient)
                .filter(Recipient.name == distributor)
                .first()
                .certs,
            )
            if not result["sigOK"]:
                LOGGER.warning(
                    "Could not verify that %s correctly signed your password entry.",
                    result["distributor"],
                )
            if not result["certOK"]:
                LOGGER.warning(
                    "Could not verify the certificate authenticity of user '%s'.",
                    result["distributor"],
                )
        oldclip = paste()
        try:
            copy(plaintext_pw)
            yield f"Password copied into paste buffer for {self.args['time']} seconds"
            sleep(self.args["time"])
        finally:
            copy(oldclip)

    def _validate_args(self):
        for argument in ["keypath"]:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

    def _pre_check(self):
        if path.exists(path.join(self.args["pwstore"], self.args["pwname"])):
            return True
        raise PasswordIOError(
            f"{path.join(self.args['pwstore'], self.args['pwname'])} does not exist"
        )



#!/usr/bin/env python3
"""This module tests the show module"""
import sys
import io
import unittest
from unittest.mock import patch
from libpkpass.commands.cli import Cli
from libpkpass.errors import DecryptionError, CliArgumentError

# from libpkpass.escrow import pk_recover_secret
from .basetest.basetest import ERROR_MSGS, patch_args


class ShowErrors(unittest.TestCase):
    """This class tests the show class"""

    def test_decryption(self):
        """Test successful decryption"""
        ret = True
        try:
            with patch_args(
                subparser_name="show",
                identity="r1",
                nopassphrase="true",
                all=None,
                pwname="test",
            ):
                "".join(Cli().run())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_recipient_not_in_database(self):
        """test what happens when the recipient is not in the appropriate directory"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(
                subparser_name="show",
                identity="bleh",
                nopassphrase="true",
                all=None,
                pwname="test",
            ):
                Cli().run()
        self.assertEqual(context.exception.msg, ERROR_MSGS["rep"])

    def test_showall_decryption(self):
        """Test showing all passwords"""
        ret = True
        try:
            with patch_args(
                subparser_name="show",
                identity="r1",
                nopassphrase="true",
                all=True,
                pwname="*test*",
            ):
                Cli().run()
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_show_nopass_error(self):
        """Test what happens when neither pwname or the all flag are supplied"""
        ret = False
        try:
            with patch_args(
                subparser_name="show", identity="r1", nopassphrase="true", all=None
            ):
                "".join(Cli().run())
        except KeyError as error:
            if str(error) == "'pwname'":
                ret = str(error)
        self.assertEqual(ret, "'pwname'")

    def test_show_recovery(self):
        try:
            shares = []
            with patch_args(
                pwname="test",
                subparser_name="show",
                identity="r2",
                nopassphrase=True,
                all=None,
                recovery="true",
            ):
                shares.append("".join(Cli().run()).split("test: ")[1])
                print(shares)
            with patch_args(
                pwname="test",
                subparser_name="show",
                identity="r3",
                nopassphrase=True,
                all=None,
                recovery="true",
            ):
                shares.append("".join(Cli().run()).split("test: ")[1])
            with patch_args(
                pwname="test",
                subparser_name="recover",
                identity="r3",
                nopassphrase=True,
            ):
                with unittest.mock.patch(
                    "builtins.input", return_value=",".join(shares)
                ):
                    passwd = "|".join(Cli().run())
        except DecryptionError as err:
            raise err
        self.assertEqual(passwd.split("|")[1], "y")


if __name__ == "__main__":
    unittest.main()
